/* zrconsole.c - main module of zrconsole
 * Copyright (C) 2002 Rik Snel <snel@phys.uu.nl>
 * Parts copied from libavcodec: 
 *            copyright (C) 2000-2002 Fabrice Bellard
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 */
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <linux/types.h>
#include <linux/videodev.h>
#include "videodev_mjpeg.h"
#define HAVE_AV_CONFIG_H
#include "libavcodec/avcodec.h"
#include "libavcodec/common.h"
#include "libavcodec/dsputil.h"
#include "libavcodec/mpegvideo.h"
#include "font_8x8.h"
#include "config2.h"

DCTELEM data[768][64];
MpegEncContext *s;
struct video_capability vc;
struct mjpeg_requestbuffers zrq;
struct mjpeg_sync zs;
struct mjpeg_params p;
typedef struct ac_code {
	UINT8 *bits;
	int size;
	int dc;
} ac_code_t;
UINT8 bits[100000];
ac_code_t ac_codes[769];

int frame = 0, synco = 0, queue = 0;
int last[768];
int cursor = 0;
unsigned char *zrbuffer;

typedef struct MJpegContext {
    UINT8 huff_size_dc_luminance[12];
    UINT16 huff_code_dc_luminance[12];
    UINT8 huff_size_dc_chrominance[12];
    UINT16 huff_code_dc_chrominance[12];

    UINT8 huff_size_ac_luminance[256];
    UINT16 huff_code_ac_luminance[256];
    UINT8 huff_size_ac_chrominance[256];
    UINT16 huff_code_ac_chrominance[256];
} MJpegContext;


static const unsigned short aanscales[64] = {
    /* precomputed values scaled up by 14 bits */
    16384, 22725, 21407, 19266, 16384, 12873,  8867,  4520,
    22725, 31521, 29692, 26722, 22725, 17855, 12299,  6270,
    21407, 29692, 27969, 25172, 21407, 16819, 11585,  5906,
    19266, 26722, 25172, 22654, 19266, 15137, 10426,  5315,
    16384, 22725, 21407, 19266, 16384, 12873,  8867,  4520,
    12873, 17855, 16819, 15137, 12873, 10114,  6967,  3552,
    8867, 12299, 11585, 10426,  8867,  6967,  4799,  2446,
    4520,  6270,  5906,  5315,  4520,  3552,  2446,  1247
};

static void convert_matrix(MpegEncContext *s, int (*qmat)[64], 
		uint16_t (*qmat16)[64], uint16_t (*qmat16_bias)[64],
		const UINT16 *quant_matrix, int bias)
{
    int qscale;

    for(qscale=1; qscale<32; qscale++){
        int i;
	if (s->fdct == ff_jpeg_fdct_islow) {
		for (i = 0; i < 64; i++) {
			const int j = s->idct_permutation[i];
			/* 16    <= qscale * quant_matrix[i] <= 7905 
			 * 19952 <= aanscales[i] *  \
			 * 	        qscale * quant_matrix[i]     <= 205026 
			 * (1<<36)/19952 >= (1<<36)/(aanscales[i] * \
			 * 	qscale * quant_matrix[i]) >= (1<<36)/249205025
			 * 3444240       >= (1<<36)/(aanscales[i] *
			 *      qscale * quant_matrix[i]) >= 275              */
			qmat[qscale][i] = (int)((UINT64_C(1) << (QMAT_SHIFT-3))/
					(qscale * quant_matrix[j]));
		}
	} else if (s->fdct == fdct_ifast) {
            for(i=0;i<64;i++) {
                const int j = s->idct_permutation[i];
                /* 16 <= qscale * quant_matrix[i] <= 7905 */
                /* 19952         <= aanscales[i] * qscale * quant_matrix[i]           <= 249205026 */
                /* (1<<36)/19952 >= (1<<36)/(aanscales[i] * qscale * quant_matrix[i]) >= (1<<36)/249205026 */
                /* 3444240       >= (1<<36)/(aanscales[i] * qscale * quant_matrix[i]) >= 275 */
                
                qmat[qscale][i] = (int)((UINT64_C(1) << (QMAT_SHIFT + 11)) / 
                                (aanscales[i] * qscale * quant_matrix[j]));
            }
        } else {
            for(i=0;i<64;i++) {
		const int j = s->idct_permutation[i];
                /* We can safely suppose that 16 <= quant_matrix[i] <= 255
                   So 16           <= qscale * quant_matrix[i]             <= 7905
                   so (1<<19) / 16 >= (1<<19) / (qscale * quant_matrix[i]) >= (1<<19) / 7905
                   so 32768        >= (1<<19) / (qscale * quant_matrix[i]) >= 67
                */
                qmat  [qscale][i] = (1 << QMAT_SHIFT_MMX) / (qscale * quant_matrix[i]);
                qmat16[qscale][i] = (1 << QMAT_SHIFT_MMX) / (qscale * quant_matrix[j]);

                if(qmat16[qscale][i]==0 || qmat16[qscale][i]==128*256) qmat16[qscale][i]=128*256-1;

                qmat16_bias[qscale][i]= ROUNDED_DIV(bias<<(16-QUANT_BIAS_SHIFT), qmat16[qscale][i]);
            }
        }
    }
}

static inline void clip_coeffs(MpegEncContext *s, DCTELEM *block, int last_index)
{
    int i;
    const int maxlevel= s->max_qcoeff;
    const int minlevel= s->min_qcoeff;

    for(i=0; i<=last_index; i++){
        const int j = s->intra_scantable.permutated[i];
        int level = block[j];
       
        if     (level>maxlevel) level=maxlevel;
        else if(level<minlevel) level=minlevel;
        block[j]= level;
    }
}

static inline void encode_dc(MpegEncContext *s, int val, 
                             UINT8 *huff_size, UINT16 *huff_code)
{
    int mant, nbits;

    if (val == 0) {
        put_bits(&s->pb, huff_size[0], huff_code[0]);
    } else {
        mant = val;
        if (val < 0) {
            val = -val;
            mant--;
        }
        
        /* compute the log (XXX: optimize) */
        nbits = 0;
        while (val != 0) {
            val = val >> 1;
            nbits++;
        }
            
        put_bits(&s->pb, huff_size[nbits], huff_code[nbits]);
        
        put_bits(&s->pb, nbits, mant & ((1 << nbits) - 1));
    }
}

static void encode_block(MpegEncContext *s, DCTELEM *block, int last_index)
{
    int mant, nbits, code, i, j, n = 0;
    int component, dc, run, /*last_index, */val;
    MJpegContext *m = s->mjpeg_ctx;
    UINT8 *huff_size_ac;
    UINT16 *huff_code_ac;
    
    /* DC coef */
    component = (n <= 3 ? 0 : n - 4 + 1);
    dc = block[0]; /* overflow is impossible */
    val = dc - s->last_dc[component];
    if (n < 4) {
        encode_dc(s, val, m->huff_size_dc_luminance, m->huff_code_dc_luminance);
        huff_size_ac = m->huff_size_ac_luminance;
        huff_code_ac = m->huff_code_ac_luminance;
    } else {
        encode_dc(s, val, m->huff_size_dc_chrominance, m->huff_code_dc_chrominance);
        huff_size_ac = m->huff_size_ac_chrominance;
        huff_code_ac = m->huff_code_ac_chrominance;
    }
    s->last_dc[component] = dc;
    
    /* AC coefs */

    run = 0;
    //last_index = last; /*s->block_last_index[n];*/
    for(i=1;i<=last_index;i++) {
        j = s->intra_scantable.permutated[i];
        val = block[j];
        if (val == 0) {
            run++;
        } else {
            while (run >= 16) {
                put_bits(&s->pb, huff_size_ac[0xf0], huff_code_ac[0xf0]);
                run -= 16;
            }
            mant = val;
            if (val < 0) {
                val = -val;
                mant--;
            }
            
            /* compute the log (XXX: optimize) */
            nbits = 0;
            while (val != 0) {
                val = val >> 1;
                nbits++;
            }
            code = (run << 4) | nbits;

            put_bits(&s->pb, huff_size_ac[code], huff_code_ac[code]);
        
            put_bits(&s->pb, nbits, mant & ((1 << nbits) - 1));
            run = 0;
        }
    }

    /* output EOB only if not already 64 values */
    if (last_index < 63 || run != 0)
        put_bits(&s->pb, huff_size_ac[0], huff_code_ac[0]);
}

int encode(unsigned char *jpg) {
	FILE *fp;
	int i, j, cursorpos, val, dc, k, i1, i2;
	unsigned char buffer[4004];
	MJpegContext *m;
	m = s->mjpeg_ctx;
	fp = fopen(VIRTUAL_TERMINAL, "r");
	fread(buffer, 1, 4004, fp);
	fclose(fp);
	//printf("%d %d %d %d\n", *buffer, *(buffer+1), *(buffer+2), *(buffer+3));
#if 0
	for (i = 0; i < 2000; i++) {
		printf("%c", buffer[i]);
	}
#endif
	init_put_bits(&s->pb, jpg, 1024*256, NULL, NULL);
	mjpeg_picture_header(s);
	if (cursor%10 > 4) {
		cursorpos = 2*(*(buffer+2)+*(buffer+3)*80) + 5;
		switch (buffer[cursorpos]) {
			case 7:
			case 15:
				buffer[cursorpos] = 112;
				break;
			case 112:
				buffer[cursorpos] = 7;
				break;
			default:
				buffer[cursorpos] = 112;
		}
	}
	
	s->header_bits = get_bit_count(&s->pb);
	s->last_dc[0] = 128;
	s->last_dc[1] = 128;
	s->last_dc[2] = 128;
	for (i = 0; i < 1000; i++) {
		switch (buffer[4*i+5]) {
			case 7:
				j = 0;
				break;
			case 15:
				j = 512;
				break;
			case 112:
				j = 256;
				break;
			default:
				j = 0;
		}
		//encode_block(s, data[buffer[4*i+4]+j], last[buffer[4*i+4]+j]);

		i1 = buffer[4*i+4]+j;
		dc = ac_codes[i1].dc;
		val = dc - s->last_dc[0];
		encode_dc(s, val, m->huff_size_dc_luminance, m->huff_code_dc_luminance);
		s->last_dc[0] = dc;
		//i1 = 33;
		//printf("%d %d %02x %02x\n", ac_codes[i1].size, ac_codes[i1].dc, ac_codes[i1].bits[0]>>4, m->huff_code_ac_luminance[0]);
		i2 = ac_codes[i1].size/8;
		//printf("--begin--");
		for (k = 0; k < i2; k++) {
		//	printf("%02x\n", ac_codes[i1].bits[k]);
			put_bits(&s->pb, 8, ac_codes[i1].bits[k]);
		}
		//printf("--eind--");
		if (ac_codes[i1].size%8 != 0) {
			put_bits(&s->pb, ac_codes[i1].size%8, ac_codes[i1].bits[i2]>>(8-(ac_codes[i1].size%8)));
		}
		//put_bits(&s->pb, m->huff_size_ac_luminance[0], m->huff_code_ac_luminance[0]);

		switch (buffer[4*i+7]) {
			case 7:
				j = 0;
				break;
			case 15:
				j = 512;
				break;
			case 112:
				j = 256;
				break;
			default:
				j = 0;
		}

		i1 = buffer[4*i+6]+j;
		dc = ac_codes[i1].dc;
		val = dc - s->last_dc[0];
		encode_dc(s, val, m->huff_size_dc_luminance, m->huff_code_dc_luminance);
		s->last_dc[0] = dc;
		

#if 1
		i2 = ac_codes[i1].size/8;
		for (k = 0; k < i2; k++) {
			put_bits(&s->pb, 8, ac_codes[i1].bits[k]);
		}
		if (ac_codes[i1].size%8 != 0) {
			put_bits(&s->pb, ac_codes[i1].size%8, ac_codes[i1].bits[i2]>>(8-(ac_codes[i1].size%8)));
		}
#endif 
#if 0
		i2 = ac_codes[i1].size/32;
		for (k = 0; k < i2; k++) {
			put_bits(&s->pb, 32, bswap_32(*((unsigned int*)ac_codes[i1].bits+k)));
		}
		if (ac_codes[i1].size%32 != 0) {
			unsigned int zooi = bswap_32(*((unsigned int*)ac_codes[i1].bits+i2));
			put_bits(&s->pb, ac_codes[i1].size%32, zooi>>(32-(ac_codes[i1].size%32)));
		}
#endif

		//encode_block(s, data[buffer[4*i+6]+j], last[buffer[4*i+6]+j]);
		/* U & V*/
		put_bits(&s->pb, 8, ac_codes[768].bits[0]);
#if 0
		put_bits(&s->pb, m->huff_size_dc_chrominance[0],
				m->huff_code_dc_chrominance[0]);
		put_bits(&s->pb, m->huff_size_ac_chrominance[0],
				m->huff_code_ac_chrominance[0]);
		put_bits(&s->pb, m->huff_size_dc_chrominance[0],
				m->huff_code_dc_chrominance[0]);
		put_bits(&s->pb, m->huff_size_ac_chrominance[0],
				m->huff_code_ac_chrominance[0]);
#endif
	}
	emms_c();
	mjpeg_picture_trailer(s);
	flush_put_bits(&s->pb);
	cursor++;
	return pbBufPtr(&s->pb) - s->pb.buf;
}

int main(int argc, char **argv) {
	int i, j, k, overflow, size, grootte, fd, ptr=0;
	int w = 1000, diff, diff2;
	struct timeval tv1, tv2;
	MJpegContext *m;
	struct timeval tv3, tv4;
	struct timeval tv5, tv6;
	char jpg[256*1024];
	for (i = 0; i < 256; i++) {
		for (j = 0; j < 8; j++) {
			for (k = 0; k < 8; k++) {
				if (fontdata_8x8[i*8+j]&128>>k) {
					data[i][8*j+k] = 127;
					data[i+256][8*j+k] = 0;
					data[i+512][8*j+k] = 255;
				} else {
					data[i][8*j+k] = 0;
					data[i+256][8*j+k] = 127;
					data[i+512][8*j+k] = 0;
				}
			}
		}
	}
	s = malloc(sizeof(MpegEncContext));
	s->width = 640;
	s->height = 200;
	s->qscale = 2;
	s->out_format = FMT_MJPEG;
	s->intra_only = 1;
	s->encoding = 1;
	s->pict_type = I_TYPE;
	s->y_dc_scale = 8;
	s->c_dc_scale = 8;
	s->mjpeg_write_tables = 1;
	s->mjpeg_vsample[0] = 1;
	s->mjpeg_vsample[1] = 1;
	s->mjpeg_vsample[2] = 1;
	s->mjpeg_hsample[0] = 2;
	s->mjpeg_hsample[1] = 1;
	s->mjpeg_hsample[2] = 1;
	avcodec_init();
	avcodec_register_all();
	if (mjpeg_init(s) < 0) exit(1);
	s->avctx = calloc(sizeof(*s->avctx), 1);
	if (MPV_common_init(s) < 0) exit(1); 
	s->mb_height = 25;
	s->mb_intra = 1;
	s->intra_matrix[0] = ff_mpeg1_default_intra_matrix[0];
	for (i = 1; i < 64; i++)
		s->intra_matrix[i] = CLAMP_TO_8BIT(
				(ff_mpeg1_default_intra_matrix[i]*s->qscale)>>3);
	convert_matrix(s, s->q_intra_matrix, s->q_intra_matrix16,
			s->q_intra_matrix16_bias,
			s->intra_matrix, s->intra_quant_bias);

	for (i = 0; i < 768; i++) {
		int last_index, mant, val, run, nbits, code;
    		UINT8 *huff_size_ac;
    		UINT16 *huff_code_ac;
    		MJpegContext *m = s->mjpeg_ctx;
		last[i] = s->dct_quantize(s, data[i], 0, 8, &overflow);
		if (overflow) clip_coeffs(s, data[i], last[i]);
		//if (i > 0) printf("blaatvoor1=%02x\n", ac_codes[0].bits[0]);
		if (((unsigned int)bits+ptr)%4 != 0) ptr += 4 - ((unsigned int)bits+ptr)%4;
		init_put_bits(&s->pb, bits+ptr, 100000-ptr, NULL, NULL);
		//if (i > 0) printf("blaatvoor2=%02x\n", ac_codes[0].bits[0]);
        	huff_size_ac = m->huff_size_ac_luminance;
        	huff_code_ac = m->huff_code_ac_luminance;
    		run = 0;
    		last_index = last[i]; /*s->block_last_index[n];*/
    		for(k=1;k<=last_index;k++) {
        		j = s->intra_scantable.permutated[k];
        		val = data[i][j];
        		if (val == 0) {
            			run++; 
			} else {
            			while (run >= 16) {
                			put_bits(&s->pb, huff_size_ac[0xf0], 
							huff_code_ac[0xf0]);
                			run -= 16;
            			}
            			mant = val;
            			if (val < 0) {
					val = -val;
                			mant--;
            			}
            
            			/* compute the log (XXX: optimize) */
           			nbits = 0;
            			while (val != 0) {
                			val = val >> 1;
					nbits++;
            			}
            			code = (run << 4) | nbits;
            			put_bits(&s->pb, huff_size_ac[code], 
						huff_code_ac[code]);
        
            			put_bits(&s->pb, nbits, 
						mant & ((1 << nbits) - 1));
            			run = 0;
        		}
    		}
    		/* output EOB only if not already 64 values */
    		if (last_index < 63 || run != 0)
        		put_bits(&s->pb, huff_size_ac[0], huff_code_ac[0]);

		ac_codes[i].bits = bits + ptr;
		ac_codes[i].size = get_bit_count(&s->pb);
		ac_codes[i].dc = data[i][0];
		flush_put_bits(&s->pb);
		ptr += pbBufPtr(&s->pb) - s->pb.buf;

//		printf("ptr=%d %d %p %d %d\n", ptr, i, ac_codes[i].bits, ac_codes[i].dc, ac_codes[i].size);
#if 0
		printf("blaatna=%02x\n", ac_codes[0].bits[0]);
		for (j = 0; j < ac_codes[i].size/8; j++) {
			printf("%02x ", ac_codes[i].bits[j]);
		}
		if (ac_codes[i].size%8 != 0)
			printf("%02x ", ac_codes[i].bits[ac_codes[i].size/8]);
		printf("\n");
#endif
	}
	if (((unsigned int)bits+ptr)%4 != 0) ptr += 4 - ((unsigned int)bits+ptr)%4;
	init_put_bits(&s->pb, bits+ptr, 100000-ptr, NULL, NULL);
	m = s->mjpeg_ctx;
	put_bits(&s->pb, m->huff_size_dc_chrominance[0],
			m->huff_code_dc_chrominance[0]);
	put_bits(&s->pb, m->huff_size_ac_chrominance[0],
			m->huff_code_ac_chrominance[0]);
	put_bits(&s->pb, m->huff_size_dc_chrominance[0],
			m->huff_code_dc_chrominance[0]);
	put_bits(&s->pb, m->huff_size_ac_chrominance[0],
			m->huff_code_ac_chrominance[0]);
	i=768;
	ac_codes[i].bits = bits + ptr;
	ac_codes[i].size = get_bit_count(&s->pb);
	ac_codes[i].dc = data[i][0];
	flush_put_bits(&s->pb);
	ptr += pbBufPtr(&s->pb) - s->pb.buf;
//	exit(1);

#if 0
	{
		FILE *fp;
		int i1, i2;
		i1 = 40;
		init_put_bits(&s->pb, jpg, 10000, NULL, NULL);

		i2 = ac_codes[i1].size/8;
		for (k = 0; k < i2; k++) {
			printf("%02x ", ac_codes[i1].bits[k]);
			put_bits(&s->pb, 8, ac_codes[i1].bits[k]);
		}
#if 0
		if (ac_codes[i1].size%8 != 0) {
			put_bits(&s->pb, ac_codes[i1].size%8, ac_codes[i1].bits[i2]>>(8-(ac_codes[i1].size%8)));
		}
#endif
		printf("%d\n", get_bit_count(&s->pb));
		fp = fopen("zooi1", "w");
		fwrite(jpg, 1, pbBufPtr(&s->pb) - s->pb.buf, fp);
		fclose(fp);

		init_put_bits(&s->pb, jpg, 10000, NULL, NULL);
		i2 = ac_codes[i1].size/16;
		for (k = 0; k < i2; k++) {
			printf("%04x ",      bswap_16(*((unsigned short*)ac_codes[i1].bits+k)));
			put_bits(&s->pb, 16, bswap_16(*((unsigned short*)ac_codes[i1].bits+k)));
		}
#if 0
		if (ac_codes[i1].size%32 != 0) {
			unsigned int zooi = bswap_32(*((unsigned int*)ac_codes[i1].bits+i2));
			put_bits(&s->pb, ac_codes[i1].size%32, zooi>>(32-(ac_codes[i1].size%32)));
		}
#endif
		printf("%d\n", get_bit_count(&s->pb));
		fp = fopen("zooi2", "w");
		fwrite(jpg, 1, pbBufPtr(&s->pb) - s->pb.buf, fp);
		fclose(fp);

	}
	exit(1);
#endif 
	fd = open(VIDEO_DEVICE, O_RDWR);
	ioctl(fd, VIDIOCGCAP, &vc);
	printf("zoran card reports: %d %d\n", vc.maxwidth, vc.maxheight);
	if (vc.maxwidth < 640 || vc.maxheight < 200) {
		printf("error 1\n");
		exit(1);
	}
	ioctl(fd, MJPIOC_G_PARAMS, &p);
	p.decimation = 0;
	p.HorDcm = 1;
	p.VerDcm = 1;
	p.TmpDcm = 1;
	p.field_per_buff = 1;
	p.img_x = (vc.maxwidth - 640)/2;
	p.img_y = (vc.maxheight/2 - 200)/2;
	p.img_width = 640;
	p.img_height = 200;
	if (ioctl(fd, MJPIOC_S_PARAMS, &p) < 0) {
		printf("error 1.5\n");
		exit(1);
	}
	zrq.count = 2;
	zrq.size = 1024*256;
	if (ioctl(fd, MJPIOC_REQBUFS, &zrq)) {
		printf("error 2\n");
		exit(1);
	}
	zrbuffer = (unsigned char*)mmap(0, zrq.count*zrq.size,
			PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);
	if (zrbuffer == MAP_FAILED) {
		printf("error 3\n");
		exit(1);
	}
	//gettimeofday(&tv1, NULL);	
	while (1) {
		//gettimeofday(&tv5, NULL);	
		grootte =encode(jpg);
		//
		//gettimeofday(&tv6, NULL);	
		if (queue - synco < 2) {
			frame = queue;
		} else {
			//gettimeofday(&tv3, NULL);	
			if (ioctl(fd, MJPIOC_SYNC, &zs) < 0) 
				printf("error waiting\n");
			//gettimeofday(&tv4, NULL);	
			diff = (tv4.tv_sec - tv3.tv_sec)*1000000 + tv4.tv_usec - tv3.tv_usec;
			if (diff > 100) w+=25;
			//printf("%d\n", diff);
			frame = zs.frame;
		//printf("frame=%d total=%d\n", frame, cursor);
			synco++;
		}
	//	gettimeofday(&tv5, NULL);	
	//	grootte =encode(zrbuffer+frame*256*1024);
	//	gettimeofday(&tv6, NULL);	
		memcpy(zrbuffer+frame*256*1024, jpg, grootte);
		diff2 = (tv6.tv_sec - tv5.tv_sec)*1000000 + tv6.tv_usec - tv5.tv_usec;
		if (ioctl(fd, MJPIOC_QBUF_PLAY, &frame) < 0) {
			printf("error queueing buffer for playback\n");
		}
		//
		//gettimeofday(&tv2, NULL);
		//printf("\rfps=%f   diff=%d grootte=%d diff2=%d  w=%d    ", (float)cursor/((float)(tv2.tv_sec-tv1.tv_sec)+ ((float)(tv2.tv_usec-tv1.tv_usec)/1000000.)), diff2, grootte, diff, w);
		//fflush(stdout);
		usleep(250);
		queue++;
	}
		
	while (queue > synco + 1) {
		if (ioctl(fd, MJPIOC_SYNC, &zs) < 0) {
			printf("error waiting for buffers to become free\n");
		}
		synco++;
	}
	frame = -1;
	if (ioctl(fd, MJPIOC_QBUF_PLAY, &frame) < 0) 
	printf("error stopping playback of last frame\n");
	close(fd);
#if 0
	size = encode(jpg);
	fp=fopen("test.jpg", "w");
	fwrite(jpg, 1, size, fp);
	fclose(fp);
#endif
	exit(0);
}
